from pydantic import BaseSettings, Field

from .client import Client


class Config(BaseSettings):

    endpoint: str = Field(
        'https://api.pollination.cloud',
        description='The API endpoint to use when making API calls',
        env='QB_POLLINATION_ENDPOINT',
    )

    token: str = Field(
        None,
        description='The API token used to login to the API',
        env='QB_POLLINATION_TOKEN'
    )

    jwt_token: str = Field(
        None,
        description='The JWT token used too authenticate to the API',
    )

    def get_client(self) -> Client:
        try:
            return Client(
                api_token=self.token,
                access_token=self.jwt_token,
                host=self.endpoint,
            )
        except ValueError as error:
            # Catch stale JWT error
            return Client(
                api_token=self.token,
                host=self.endpoint,
            )



import os
from concurrent.futures import ThreadPoolExecutor as PoolExecutor
from multiprocessing import cpu_count
import tarfile
from urllib3.exceptions import ProtocolError

import requests
from tabulate import tabulate

from queenbee.workflow import Arguments
from queenbee.workflow.status import WorkflowStatus

from pollination_sdk.exceptions import ApiException
from pollination_sdk import models

from ..client import Client

try:
    import click
except ImportError:
    raise ImportError(
        'click modules not installed. Try `pip install queenbee-pollination[cli]` command.'
    )


def handle_project(client: Client, owner: str, name: str):

    try:
        client.projects.get_project(
            owner=owner,
            name=name,
        )
    except ApiException as error:
        if error.status != 404:
            raise click.ClickException(error)

        if click.confirm(
            f'Project not found. Do you want to create a new project at {owner}/{name}',
            abort=True
        ):
            private = click.confirm('Do you want to create a private project?'),

            public = True

            if private is True:
                public = False

            new_project = models.PatchProjectDto(
                name=name,
                public=public,
                description=''
            )

            try:
                client.projects.create_project(
                    owner=owner,
                    patch_project_dto=new_project,
                )
            except ApiException as error:
                raise click.ClickException(error)

            click.echo('Successfully created project!')


@click.group('project')
def project():
    pass


@project.group('folder')
def folder():
    pass

@folder.command('upload')
@click.argument('path', default='.', type=click.Path(exists=True))
@click.option('-p', '--project', type=str, required=True)
@click.option('-o', '--owner', help='a pollination account name')
def upload_folder(path, owner, project):
    """upload project files"""

    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    handle_project(
        client=client,
        owner=owner,
        name=project,
    )


    def _upload_artifact(key: str):
        res = client.artifacts.create_artifact(
            owner=owner,
            name=project,
            key_request={'key': key.replace('\\', '/')}
        )

        # Demonstrate how another Python program can use the presigned URL to upload a file
        with open(key, 'rb') as f:
            files = {'file': (key, f)}
            http_response = requests.post(
                res.url, data=res.fields, files=files)

            if http_response.status_code == 204:
                click.echo(f"Uploaded {key}")


    keys = []
    for root, subdirs, files in os.walk(path):
        for fi in files:
            key = os.path.join(root, fi)
            keys.append(key)

    worker_count = max(cpu_count() - 1, 1)
    with PoolExecutor(max_workers=worker_count) as executor:
        for _ in executor.map(_upload_artifact, keys):
            pass


@folder.command('delete')
@click.option('-p', '--project', help='project name', type=str, required=True)
@click.option('-o', '--owner', help='a pollination account name')
@click.option('--path', help='the subpath/subfolder to delete')
def delete_artifacts(project, owner, path):
    """delete project files"""

    if path is not None:
        path = [path]

    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    client.artifacts.delete_artifact(
        owner=owner,
        name=project,
        path=path
    )

    print("Poof... All gone!")


@project.group('simulation')
def simulation():
    pass


@simulation.command('submit')
@click.argument('recipe', type=str)
@click.option('-p', '--project', help='project name', type=str, required=True)
@click.option('-o', '--owner', help='a pollination account name')
@click.option('-i', '--inputs', help='path to an inputs file', type=click.Path(exists=True))
def submit(project, recipe, owner, inputs):
    """Schedule a simulation to be run"""

    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    handle_project(
        client=client,
        owner=owner,
        name=project,
    )

    try:
        rec_owner, rec_name = recipe.split('/')
        rec_name, rec_tag = rec_name.split(':')
    except ValueError:
        click.ClickException(f'Expected recipe reference in format "owner/name:tag" not: {recipe}')

    recipe_ref = models.RecipeSelection(
        owner=rec_owner,
        name=rec_name,
        tag=rec_tag,
    )

    arguments = Arguments()

    if inputs is not None:
        arguments = Arguments.from_file(inputs)

    submit = models.SubmitSimulationDto(
        recipe=recipe_ref,
        inputs=arguments.to_dict()
    )

    try:
        res = client.simulations.create_simulation(
            owner=owner,
            name=project,
            submit_simulation_dto=submit,
        )
    except ApiException as error:
        raise click.ClickException(error)

    click.echo(f'Successfully scheduled simulation: {res.id}')



@simulation.command('list')
@click.option('-p', '--project', type=str, required=True)
@click.option('-o', '--owner', help='a pollination account name')
@click.option('--page', type=int, default=1, show_default=True)
def list_simulations(project, owner, page):
    """List simulations for a given project"""

    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    try:
        res = client.simulations.list_simulations(
            owner=owner,
            name=project,
            page=page,
        )
    except ApiException as error:
        raise click.ClickException(error)

    table = [[sim.id, sim.status, sim.started_at, sim.finished_at] for sim in res.resources]

    table.sort(key=lambda r: r[2], reverse=True)

    print(tabulate(table, headers=['ID', 'Status', 'Stated At', 'Finished At']))


@simulation.command('download')
@click.option('-p', '--project', help='Name of a project', required=True)
@click.option('-i', '--id', help='ID of the simulation your want to retrieve artifacts from', required=True)
@click.option('-o', '--owner', help='a pollination account name')
@click.option('-f', '--folder', help='Folder path to save simulation artifacts to', default='.', show_default=True)
@click.option('-a', '--artifact', help='Choose which artifacts to download. Will download all if not specified.', type=click.Choice(['inputs', 'outputs', 'logs'], case_sensitive=False))
def download_simulation_artifacts(owner, project, id, folder, artifact):
    """download simulation artifacts"""

    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    folder = os.path.join(folder, id)

    if artifact is None:
        artifacts = ['inputs', 'outputs', 'logs']
    else:
        artifacts = [artifact]

    try:
        simulation = client.simulations.get_simulation(
            owner=owner,
            name=project,
            simulation_id=id,
        )
    except ApiException as error:
        if error.status == 404:
            raise click.ClickException(
                f'Simulation not found: {owner}/{project}/{id}')
        raise click.ClickException(error)

    for a in artifacts:

        if a == 'inputs':
            try:
                fetch_url = client.simulations.get_simulation_inputs(
                    owner=owner,
                    name=project,
                    simulation_id=id,
                )
            except ApiException as error:
                raise click.ClickException(error)

        elif a == 'outputs':
            try:
                fetch_url = client.simulations.get_simulation_outputs(
                    owner=owner,
                    name=project,
                    simulation_id=id,
                )
            except ApiException as error:
                raise click.ClickException(error)

        elif a == 'logs':
            try:
                fetch_url = client.simulations.get_simulation_logs(
                    owner=owner,
                    name=project,
                    simulation_id=id,
                )
            except ApiException as error:
                raise click.ClickException(error)

        response = requests.get(url=fetch_url)

        try:
            with open(f'{folder}.tar.gz', 'wb') as f:
                f.write(response.content)

            try:
                tar = tarfile.open(name=f'{folder}.tar.gz')
                tar.extractall(f'{folder}')
                click.echo(f'Saved {a} files to {folder}/{a}')
            except tarfile.ReadError as error:
                click.echo(f'Failed to read {a}')
            os.remove(f'{folder}.tar.gz')

        except ProtocolError:
            click.echo(f'No {a}  files found')

    status = WorkflowStatus.parse_obj(simulation.to_dict())

    status.to_yaml(
        filepath=os.path.join(folder, 'simulation.yml')
    )



import os
import shutil

from queenbee.recipe import Recipe
from queenbee.operator import Operator
from queenbee.repository.package import RecipeVersion, OperatorVersion

from pollination_sdk.exceptions import ApiException

try:
    import click
except ImportError:
    raise ImportError(
        'click modules not installed. Try `pip install queenbee-pollination[cli]` command.'
    )


@click.group('pull')
def pull():
    pass


@pull.command('recipe')
@click.argument('name')
@click.option('-o', '--owner', help='a pollination account name')
@click.option('-t', '--tag', help='specific recipe tag to pull')
@click.option('-p', '--path', help='path to download the recipe to', type=click.Path(exists=True), default='.')
@click.option('-f', '--force', help='overwrite local files', type=bool, default=False, is_flag=True)
def recipe(name, owner, tag, path, force):
    """download a recipe from a pollination registry"""
    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    path = os.path.abspath(path)
    path = os.path.join(path, name)

    if owner is None:
        account = client.get_account()
        owner = account.username

    if tag is None:
        try:
            res = client.recipes.get_recipe(
                owner=owner,
                name=name,
            )
        except ApiException as error:
            if error.status == 404:
                raise click.ClickException(f'Recipe not found: {owner}/{name}')
            raise click.ClickException(error)

        tag = res['latest_tag']

    try:
        res = client.recipes.get_recipe_tag(
            owner=owner,
            name=name,
            tag=tag
        )
    except ApiException as error:
        if error.status == 404:
            raise click.ClickException(f'Recipe not found: {owner}/{name}:{tag}')
        raise click.ClickException(error)

    manifest = Recipe.parse_obj(res['manifest'])

    try:
        manifest.to_folder(
            folder_path=path,
            readme_string=res['readme'],
            license_string=res['license'],
        )
    except FileExistsError as error:
        if not force:
            raise click.ClickException(f'Folder already exists at path {path}. Use "--force" to overwrite it.')

        shutil.rmtree(path)

        manifest.to_folder(
            folder_path=path,
            readme_string=res['readme'],
            license_string=res['license'],
        )

    click.echo(f'Recipe {owner}/{name}:{tag} saved to {path}')


@pull.command('operator')
@click.argument('name')
@click.option('-o', '--owner', help='a pollination account name')
@click.option('-t', '--tag', help='specific recipe tag to pull')
@click.option('-p', '--path', help='path to download the operator to', type=click.Path(exists=True), default='.')
@click.option('-f', '--force', help='overwrite local files', type=bool, default=False, is_flag=True)
def operator(name, owner, tag, path, force):
    """download an operator from a pollination registry"""

    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    path = os.path.abspath(path)
    path = os.path.join(path, name)

    if owner is None:
        account = client.get_account()
        owner = account.username

    if tag is None:
        try:
            res = client.operators.get_operator(
                owner=owner,
                name=name,
            )
        except ApiException as error:
            if error.status == 404:
                raise click.ClickException(f'Operator not found: {owner}/{name}')
            raise click.ClickException(error)

        tag = res['latest_tag']

    try:
        res = client.operators.get_operator_tag(
            owner=owner,
            name=name,
            tag=tag
        )
    except ApiException as error:
        if error.status == 404:
            raise click.ClickException(
                f'Operator not found: {owner}/{name}:{tag}')
        raise click.ClickException(error)

    manifest = Operator.parse_obj(res['manifest'])

    try:
        manifest.to_folder(
            folder_path=path,
            readme_string=res['readme'],
            license_string=res['license'],
        )
    except FileExistsError:
        if not force:
            raise click.ClickException(f'Folder already exists at path {path}')

        shutil.rmtree(path)

        manifest.to_folder(
            folder_path=path,
            readme_string=res['readme'],
            license_string=res['license'],
        )

    click.echo(f'Operator {owner}/{name}:{tag} saved to {path}')



from pydantic import ValidationError

from queenbee.recipe import Recipe
from queenbee.operator import Operator
from queenbee.repository.package import RecipeVersion, OperatorVersion

from pollination_sdk.exceptions import ApiException
from pollination_sdk.models import NewRepositoryDto

from ..client import Client

try:
    import click
except ImportError:
    raise ImportError(
        'click modules not installed. Try `pip install queenbee-pollination[cli]` command.'
    )


def handle_repository(client: Client, repo_type: str, owner: str, name: str, create_repo: bool = False):

    if repo_type not in ['recipe', 'operator']:
        raise click.ClickException('Repository type should be one of ["recipe", "operator"]')

    try:
        if repo_type == 'recipe':
            client.recipes.get_recipe(
                owner=owner,
                name=name,
            )
        elif repo_type == 'operator':
            client.operators.get_operator(
                owner=owner,
                name=name,
            )
    except ApiException as error:
        if error.status != 404:
            raise click.ClickException(error)

        if create_repo:
            new_repo = NewRepositoryDto(
                public=True,
                name=name,
            )

        else:
            if click.confirm(
                f'{repo_type} repository not found. Do you want to create a new repository at {owner}/{name}',
                abort=True
            ):
                private = click.confirm(
                    'Do you want to create a private repository?'),

                public = True

                if private is True:
                    public = False

                new_repo = NewRepositoryDto(
                    public=public,
                    name=name,
                )

        try:
            if repo_type == 'recipe':
                client.recipes.create_recipe(
                    owner=owner,
                    new_repository_dto=new_repo,
                )
            elif repo_type == 'operator':
                client.operators.create_operator(
                    owner=owner,
                    new_repository_dto=new_repo,
                )
        except ApiException as error:
            raise click.ClickException(error)

        click.echo('Successfully created repository!')


@click.group('push')
def push():
    pass


@push.command('recipe')
@click.argument('path', type=click.Path(exists=True))
@click.option('-o', '--owner', help='a pollination account name')
@click.option('-t', '--tag', help='tag to apply to the recipe')
@click.option('--create-repo', help='create the recipe repository if it does not exist (defaults to a public repo)', type=bool, default=False, is_flag=True)
def recipe(path, owner, tag, create_repo):
    """push a queenbee recipe to the pollination registry

    This subcommand pushes a packaged queenbee recipe to a registry on
    pollination cloud
    """
    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    try:
        manifest = Recipe.from_folder(path)
    except ValidationError as error:
        raise click.ClickException(error)
    except FileNotFoundError as error:
        raise click.ClickException(error)
    except Exception as error:
        raise error

    handle_repository(
        client=client,
        repo_type='recipe',
        owner=owner,
        name=manifest.metadata.name,
        create_repo=create_repo
    )

    if tag is not None:
        manifest.metadata.tag = tag

    readme_string = RecipeVersion.read_readme(path)
    license_string = RecipeVersion.read_license(path)

    if readme_string is None:
        readme_string = ''

    if license_string is None:
        license_string = ''

    new_package = {
        'manifest': manifest.to_dict(),
        'readme': readme_string,
        'license': license_string
    }

    try:
        client.recipes.create_recipe_package(
            owner=owner,
            name=manifest.metadata.name,
            new_recipe_package=new_package,
        )
    except ApiException as error:
        raise click.ClickException(error)

    click.echo(f'Successfully created new recipe package {owner}/{manifest.metadata.name}:{manifest.metadata.tag}')


@push.command('operator')
@click.argument('path', type=click.Path(exists=True))
@click.option('-o', '--owner', help='a pollination account name')
@click.option('-t', '--tag', help='tag to apply to the operator')
@click.option('--create-repo', help='create the operator repository if it does not exist (defaults to a public repo)', type=bool, default=False, is_flag=True)
def operator(path, owner, tag, create_repo):
    """push a queenbee operator to the pollination registry

    This subcommand pushes a packaged queenbee operator to a registry on
    pollination cloud
    """
    ctx = click.get_current_context()
    client = ctx.obj.get_client()

    if owner is None:
        account = client.get_account()
        owner = account.username

    try:
        manifest = Operator.from_folder(path)
    except ValidationError as error:
        raise click.ClickException(error)
    except FileNotFoundError as error:
        raise click.ClickException(error)
    except Exception as error:
        raise error

    handle_repository(
        client=client,
        repo_type='operator',
        owner=owner,
        name=manifest.metadata.name,
        create_repo=create_repo
    )

    if tag is not None:
        manifest.metadata.tag = tag

    readme_string = RecipeVersion.read_readme(path)
    license_string = RecipeVersion.read_license(path)

    if readme_string is None:
        readme_string = ''

    if license_string is None:
        license_string = ''

    new_package = {
        'manifest': manifest.to_dict(),
        'readme': readme_string,
        'license': license_string
    }

    try:
        client.operators.create_operator_package(
            owner=owner,
            name=manifest.metadata.name,
            new_operator_package=new_package,
        )
    except ApiException as error:
        raise click.ClickException(error)

    click.echo(
        f'Successfully created new operator package {owner}/{manifest.metadata.name}:{manifest.metadata.tag}')



import pollination_sdk as sdk


class Client(object):
    """A Pollination client designed to interact with Workflow and Simulation objects."""

    def __init__(self, api_token=None, access_token=None, host='https://api.pollination.cloud'):
        config = sdk.Configuration()
        config.host = host

        auth = sdk.UserApi(sdk.ApiClient(config))

        if access_token is None:
            api_token = sdk.LoginDto(
                api_token=api_token
            )

            try:
                auth_response = auth.login(api_token)
            except sdk.rest.ApiException as error:
                if error.status == 403:
                    raise ValueError("Failed to log in... </3")
                raise error

            config.access_token = auth_response.access_token
        else:
            config.access_token = access_token

        self.config = config
        self.auth = sdk.UserApi(sdk.ApiClient(config))
        self.recipes = sdk.RecipesApi(sdk.ApiClient(config))
        self.operators = sdk.OperatorsApi(sdk.ApiClient(config))
        self.simulations = sdk.SimulationsApi(sdk.ApiClient(config))
        self.artifacts = sdk.ArtifactsApi(sdk.ApiClient(config))
        self.projects = sdk.ProjectsApi(sdk.ApiClient(config))

    def get_account(self) -> sdk.models.PrivateUserDto:
        return self.auth.get_me()



try:
    import click
except ImportError:
    raise ImportError(
        'click modules is not installed. Try `pip install queenbee[cli]` command.'
    )

from .context import Context
from .pull import pull
from .push import push
from .project import project

@click.group(invoke_without_command=True)
@click.version_option()
def pollination():
    """
    pollination cloud plugin
    """
    ctx = click.get_current_context()
    queenbee_config = ctx.obj

    ctx.ensure_object(Context)
    ctx.obj.queenbee = queenbee_config

    if ctx.invoked_subcommand is None:
        click.echo(ctx.command.get_help(ctx))


pollination.add_command(pull)
pollination.add_command(push)
pollination.add_command(project)


import setuptools

with open('README.md') as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('cli-requirements.txt') as f:
    cli_requirements = f.read().splitlines()

setuptools.setup(
    name="queenbee_pollination",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="Ladybug Tools",
    author_email="info@ladybug.tools",
    description="queenbee-pollination extends queenbee to interact with the Pollination API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pollination/queenbee-pollination",
    packages=setuptools.find_packages(exclude=["tests", "hack", "docs"]),
    install_requires=requirements,
    extras_require={'cli': cli_requirements},
    entry_points='''
        [queenbee.plugins]
        pollination=queenbee_pollination.cli:pollination
    ''',
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent"
    ],
)



from pydantic import Field
from queenbee.base.basemodel import BaseModel
from queenbee.config import Config as QueenbeeConfig

from ..config import Config as QueenbeePollinationConfig
from ..client import Client


class Context(BaseModel):

    queenbee: QueenbeeConfig = Field(
        None,
        description='The queenbee config object'
    )

    config: QueenbeePollinationConfig = Field(
        default_factory=QueenbeePollinationConfig,
    )

    def get_client(self) -> Client:
        self.queenbee.refresh_tokens()
        auth_header = self.queenbee.config.get_auth_header(
            registry_url=self.config.endpoint
        )

        if auth_header is not None:
            self.config.jwt_token = auth_header.split('Bearer ')[-1]
        return self.config.get_client()
